
#include <stdio.h>
#include <math.h>
#include "nr_util.h"
#include "ad_globl.h"
#include "ad_frsnl.h"
#include "iad_type.h"
#include "iad_util.h"
#include "iad_find.h"
#include "iad_pub.h"
#include "iad_io.h"
#include "mc_lost.h"
#define DEBUG_SEARCH 0



void
Inverse_RT (struct measure_type m, struct invert_type *r)
{
  r->found = FALSE;

  r->error = measure_OK (m, *r);
  if (r->method.quad_pts < 4)
    r->error = IAD_QUAD_PTS_NOT_VALID;

  if (r->error != IAD_NO_ERROR)
    return;



  r->search = determine_search (m, *r);


  switch (r->search)
    {
    case FIND_A:
      U_Find_A (m, r);
      break;
    case FIND_B:
      U_Find_B (m, r);
      break;
    case FIND_G:
      U_Find_G (m, r);
      break;
    case FIND_Ba:
      U_Find_Ba (m, r);
      break;
    case FIND_Bs:
      U_Find_Bs (m, r);
      break;

    case FIND_AB:
      U_Find_AB (m, r);
      break;
    case FIND_AG:
      U_Find_AG (m, r);
      break;
    case FIND_BG:
      U_Find_BG (m, r);
      break;
    case FIND_BsG:
      U_Find_BsG (m, r);
      break;
    case FIND_BaG:
      U_Find_BaG (m, r);
      break;
    }
  if (r->iterations == IAD_MAX_ITERATIONS)
    r->error = IAD_TOO_MANY_ITERATIONS;


  if (r->error == IAD_NO_ERROR && r->final_distance <= r->tolerance)
    r->found = TRUE;
}




int
measure_OK (struct measure_type m, struct invert_type r)
{
  double rt, tt, rd, rc, td, tc;

  int error = IAD_NO_ERROR;



  if (m.m_r < 0)
    error = IAD_R_LT_ZERO;
  if (m.m_r > 1)
    error = IAD_R_GT_ONE;


  if (m.num_measures > 1)
    {
      if (m.m_t < 0)
	error = IAD_T_LT_ZERO;
      if (m.m_t > 1)
	error = IAD_T_GT_ONE;
      if (m.m_t + m.m_r > 1)
	error = IAD_R_PLUS_T_GT_ONE;
    }


  if (m.num_measures == 3)
    {
      if (m.m_u > 1)
	error = IAD_TU_GT_ONE;
      if (m.m_u < 0)
	error = IAD_TU_LT_ZERO;
      if (!m.sphere_with_tc && m.m_r + m.m_t + m.m_u > 1.0)
	error = IAD_R_PLUS_T_PLUS_TU_GT_ONE;
    }




  Estimate_RT (m, r.slab, &rt, &tt, &rd, &rc, &td, &tc);



  if (rd < 0)
    error = IAD_RD_LT_ZERO;
  if (rt > 1.0)
    error = IAD_RT_GT_ONE;
  if (td < 0)
    error = IAD_TD_LT_ZERO;
  if (tt > 1.0)
    error = IAD_TT_GT_ONE;
  if (tt + rt > 1.0)
    error = IAD_RT_PLUS_TT_GT_ONE;


  if (m.num_measures > 1)
    {
      double rmin, tmin, b;

      if (m.m_u == 0)
	b = What_Is_B (r.slab, m.m_t);
      else
	b = What_Is_B (r.slab, m.m_u);

      Sp_mu_RT (r.slab.n_top_slide, r.slab.n_slab, r.slab.n_bottom_slide,
		r.slab.b_top_slide, b, r.slab.b_bottom_slide, 1.0, &rmin,
		&tmin);

      if (m.m_r + 0.0001 < rmin)
	{
	  error = IAD_RT_LT_MINIMUM;
	  fprintf (stderr,
		   "Not enough reflected light for this measurement\n");
	  fprintf (stderr,
		   "your    reflectance = %7.5f  transmittance = %7.5f \n",
		   m.m_r, m.m_t);
	  fprintf (stderr, "minimum reflectance = %7.5f\n", rmin);
	}
    }



  if (rd == 0 && td != 0)
    if (r.search != FIND_Ba && r.search != FIND_Bs)
      error = IAD_RD_IS_ZERO_BUT_NOT_TD;





  if (m.num_spheres != 0)
    {

      if (m.as_r < 0 || m.as_r >= 0.2)
	error = IAD_AS_NOT_VALID;

      if (m.ad_r < 0 || m.ad_r >= 0.2)
	error = IAD_AD_NOT_VALID;

      if (m.ae_r < 0 || m.ae_r >= 0.2)
	error = IAD_AE_NOT_VALID;

      if (m.rw_r < 0 || m.rw_r > 1.0)
	error = IAD_RW_NOT_VALID;

      if (m.rd_r < 0 || m.rd_r > 1.0)
	error = IAD_RD_NOT_VALID;

      if (m.rstd_r < 0 || m.rstd_r > 1.0)
	error = IAD_RSTD_NOT_VALID;

      if (m.f_r < 0 || m.f_r > 1)
	error = IAD_F_NOT_VALID;
    }



  if (m.num_spheres != 0)
    {

      if (m.as_t < 0 || m.as_t >= 0.2)
	error = IAD_AS_NOT_VALID;

      if (m.ad_t < 0 || m.ad_t >= 0.2)
	error = IAD_AD_NOT_VALID;

      if (m.ae_t < 0 || m.ae_t >= 0.2)
	error = IAD_AE_NOT_VALID;

      if (m.rw_t < 0 || m.rw_r > 1.0)
	error = IAD_RW_NOT_VALID;

      if (m.rd_t < 0 || m.rd_t > 1.0)
	error = IAD_RD_NOT_VALID;

      if (m.rstd_t < 0 || m.rstd_t > 1.0)
	error = IAD_RSTD_NOT_VALID;

      if (m.f_t < 0 || m.f_t > 1)
	error = IAD_F_NOT_VALID;
    }


  return error;
}




search_type
determine_search (struct measure_type m, struct invert_type r)
{
  double rt, tt, rd, td, tc, rc;
  int search = 0;
  int independent = m.num_measures;

  Estimate_RT (m, r.slab, &rt, &tt, &rd, &rc, &td, &tc);

  if (tc == 0 && independent == 3)
    independent--;

  if (rd == 0 && independent == 2)
    independent--;

  if (td == 0 && independent == 2)
    independent--;

  if (independent == 1)
    {

      if (r.default_a != UNINITIALIZED)
	{
	  if (td == 0)
	    search = FIND_G;
	  else
	    search = FIND_B;
	}

      else if (r.default_b != UNINITIALIZED)
	search = FIND_A;

      else if (r.default_bs != UNINITIALIZED)
	search = FIND_Ba;

      else if (r.default_ba != UNINITIALIZED)
	search = FIND_Bs;

      else if (rd != 0 || td == 0)
	search = FIND_A;
      else
	search = FIND_B;



    }

  else if (independent == 2)
    {

      if (r.default_a != UNINITIALIZED)
	search = FIND_BG;

      else if (r.default_b != UNINITIALIZED)
	search = FIND_AG;

      else if (r.default_ba != UNINITIALIZED)
	search = FIND_BsG;

      else if (r.default_bs != UNINITIALIZED)
	search = FIND_BaG;

      else
	search = FIND_AB;


    }


  else
    {
      search = FIND_AG;
    }

  if (DEBUG_SEARCH)
    {
      fprintf (stderr, "FIND_A    = 0\n");
      fprintf (stderr, "FIND_B    = 1\n");
      fprintf (stderr, "FIND_AB   = 2\n");
      fprintf (stderr, "FIND_AG   = 3\n");
      fprintf (stderr, "FIND_AUTO = 4\n");
      fprintf (stderr, "FIND_BG   = 5\n");
      fprintf (stderr, "FIND_BaG  = 6\n");
      fprintf (stderr, "FIND_BsG  = 7\n");
      fprintf (stderr, "FIND_Ba   = 8\n");
      fprintf (stderr, "FIND_Bs   = 9\n");
      fprintf (stderr, "FIND_G    =10\n");

      fprintf (stderr, "search    =%2d\n", search);
    }

  return search;
}





void
Initialize_Result (struct measure_type m, struct invert_type *r)
{

  r->a = 0.0;
  r->b = 0.0;
  r->g = 0.0;


  r->found = FALSE;
  r->tolerance = 0.0001;
  r->MC_tolerance = 0.01;
  r->search = FIND_AUTO;
  r->metric = RELATIVE;
  r->final_distance = 10;
  r->iterations = 0;
  r->error = IAD_NO_ERROR;


  r->default_a = UNINITIALIZED;
  r->default_b = UNINITIALIZED;
  r->default_g = UNINITIALIZED;
  r->default_ba = UNINITIALIZED;
  r->default_bs = UNINITIALIZED;




  r->slab.a = 0.5;
  r->slab.b = 1.0;
  r->slab.g = 0;
  r->slab.phase_function = HENYEY_GREENSTEIN;
  r->slab.n_slab = m.slab_index;
  r->slab.n_top_slide = m.slab_top_slide_index;
  r->slab.n_bottom_slide = m.slab_bottom_slide_index;
  r->slab.b_top_slide = m.slab_top_slide_b;
  r->slab.b_bottom_slide = m.slab_bottom_slide_b;

  r->method.a_calc = 0.5;
  r->method.b_calc = 1;
  r->method.g_calc = 0.5;
  r->method.quad_pts = 8;
  r->method.b_thinnest = 1.0 / 32.0;


}




void
Initialize_Measure (struct measure_type *m)
{
  double default_sphere_d = 8.0 * 25.4;
  double default_sample_d = 0.5 * 25.4;
  double default_detector_d = 0.1 * 25.4;
  double default_entrance_d = 0.5 * 25.4;
  double sphere = default_sphere_d * default_sphere_d;

  m->slab_index = 1.0;
  m->slab_top_slide_index = 1.0;
  m->slab_top_slide_b = 0.0;
  m->slab_top_slide_thickness = 0.0;
  m->slab_bottom_slide_index = 1.0;
  m->slab_bottom_slide_b = 0.0;
  m->slab_bottom_slide_thickness = 0.0;
  m->slab_thickness = 1.0;

  m->num_spheres = 0;
  m->num_measures = 1;

  m->sphere_with_rc = 1.0;
  m->sphere_with_tc = 1.0;

  m->m_r = 0.0;
  m->m_t = 0.0;
  m->m_u = 0.0;

  m->d_sphere_r = default_sphere_d;
  m->as_r = default_sample_d * default_sample_d / sphere;
  m->ad_r = default_detector_d * default_detector_d / sphere;
  m->ae_r = default_entrance_d * default_entrance_d / sphere;
  m->aw_r = 1.0 - m->as_r - m->ad_r - m->ae_r;
  m->rd_r = 0.0;
  m->rw_r = 1.0;
  m->rstd_r = 1.0;
  m->f_r = 0.0;

  m->d_sphere_t = default_sphere_d;
  m->as_t = m->as_r;
  m->ad_t = m->ad_r;
  m->ae_t = m->ae_r;
  m->aw_t = m->aw_r;
  m->rd_t = 0.0;
  m->rw_t = 1.0;
  m->rstd_t = 1.0;
  m->f_t = 0.0;

  m->lambda = 0.0;
  m->d_beam = 0.0;
  m->ur1_lost = 0;
  m->uru_lost = 0;
  m->ut1_lost = 0;
  m->utu_lost = 0;

}




void
ez_Inverse_RT (double n, double nslide, double UR1, double UT1, double Tc,
	       double *a, double *b, double *g, int *error)
{
  struct measure_type m;
  struct invert_type r;
  *a = 0;
  *b = 0;
  *g = 0;

  Initialize_Measure (&m);

  m.slab_index = n;
  m.slab_top_slide_index = nslide;
  m.slab_bottom_slide_index = nslide;

  m.num_measures = 3;
  if (UT1 == 0)
    m.num_measures--;
  if (Tc == 0)
    m.num_measures--;

  m.m_r = UR1;
  m.m_t = UT1;
  m.m_u = Tc;

  Initialize_Result (m, &r);
  r.method.quad_pts = 8;

  Inverse_RT (m, &r);

  *error = r.error;
  if (r.error == IAD_NO_ERROR)
    {
      *a = r.a;
      *b = r.b;
      *g = r.g;
    }
}




void
Spheres_Inverse_RT (double *setup,
		    double *analysis,
		    double *sphere_r,
		    double *sphere_t, double *measurements, double *results)
{
  struct measure_type m;
  struct invert_type r;
  long num_photons;
  double ur1, ut1, uru, utu;
  int i, mc_runs = 1;

  Initialize_Measure (&m);


  {
    double d_sample_r, d_entrance_r, d_detector_r;
    double d_sample_t, d_entrance_t, d_detector_t;

    m.slab_index = setup[0];
    m.slab_top_slide_index = setup[1];
    m.slab_thickness = setup[2];
    m.slab_top_slide_thickness = setup[3];
    m.d_beam = setup[4];
    m.rstd_r = setup[5];
    m.num_spheres = (int) setup[6];

    m.d_sphere_r = setup[7];
    d_sample_r = setup[8];
    d_entrance_r = setup[9];
    d_detector_r = setup[10];
    m.rw_r = setup[11];

    m.d_sphere_t = setup[12];
    d_sample_t = setup[13];
    d_entrance_t = setup[14];
    d_detector_t = setup[15];
    m.rw_t = setup[16];

    r.default_g = setup[17];
    num_photons = (long) setup[18];

    m.as_r = (d_sample_r / m.d_sphere_r) * (d_sample_r / m.d_sphere_r);
    m.ae_r = (d_entrance_r / m.d_sphere_r) * (d_entrance_r / m.d_sphere_r);
    m.ad_r = (d_detector_r / m.d_sphere_r) * (d_detector_r / m.d_sphere_r);
    m.aw_r = 1.0 - m.as_r - m.ae_r - m.ad_r;
    m.as_t = (d_sample_t / m.d_sphere_t) * (d_sample_t / m.d_sphere_t);
    m.ae_t = (d_entrance_t / m.d_sphere_t) * (d_entrance_t / m.d_sphere_t);
    m.ad_t = (d_detector_t / m.d_sphere_t) * (d_detector_t / m.d_sphere_t);
    m.aw_t = 1.0 - m.as_t - m.ae_t - m.ad_t;

    m.slab_bottom_slide_index = m.slab_top_slide_index;
    m.slab_bottom_slide_thickness = m.slab_top_slide_thickness;
  }



  m.as_r = sphere_r[0];
  m.ae_r = sphere_r[1];
  m.ad_r = sphere_r[2];
  m.rw_r = sphere_r[3];
  m.rd_r = sphere_r[4];
  m.rstd_r = sphere_r[5];
  m.f_r = sphere_r[7];



  m.as_t = sphere_t[0];
  m.ae_t = sphere_t[1];
  m.ad_t = sphere_t[2];
  m.rw_t = sphere_t[3];
  m.rd_t = sphere_t[4];
  m.rstd_t = sphere_t[5];
  m.f_t = sphere_t[7];


  m.m_r = measurements[0];
  m.m_t = measurements[1];
  m.m_u = measurements[2];

  m.num_measures = 3;
  if (m.m_t == 0)
    m.num_measures--;
  if (m.m_u == 0)
    m.num_measures--;



  Initialize_Result (m, &r);
  results[0] = 0;
  results[1] = 0;
  results[2] = 0;



  r.method.quad_pts = (int) analysis[0];
  mc_runs = (int) analysis[1];



  Inverse_RT (m, &r);
  for (i = 0; i < mc_runs; i++)
    {
      MC_Lost (m, r, num_photons, &ur1, &ut1, &uru, &utu,
	       &m.ur1_lost, &m.ut1_lost, &m.uru_lost, &m.utu_lost);
      Inverse_RT (m, &r);
    }

  if (r.error == IAD_NO_ERROR)
    {
      results[0] = (1 - r.a) * r.b / m.slab_thickness;
      results[1] = (r.a) * r.b / m.slab_thickness;
      results[2] = r.g;
    }

  results[3] = r.error;
}
